/**
 * Copyright (c) 2006-2009 Institut National de Recherche en Informatique et 
 *                         en Automatique. All rights reserved.
 *
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */

#include <stdlib.h>
#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h> 
#include <string.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "stubs/serviceH.h"
#include "stubs/service.nsmap"
#include "broker.h"
#include "debug.h"
#include "cmdline.h"

struct Namespace *namespaces;

static struct gengetopt_args_info ARGS;
static struct cluster *clusters = NULL;
static int clusters_count = 0;
static pthread_mutex_t clusters_lock = PTHREAD_MUTEX_INITIALIZER;

static int  running = 1;

static int getConfiguration();
static int freeClusters();
static int acceptFrontalesConnections();
static void closeAllSockets();
static void setup_signals(void);
static void* handleRequest(void *p_tsoap);
static void debugClusters(void);
static int doconnect(struct soap *soap, const char *endpoint, const char *host, int port);
static int doclose(struct soap *soap);


pthread_mutex_t frontale_lock = PTHREAD_MUTEX_INITIALIZER;

int CONNECTED_SOCKET = 0;

int main(int argc, char *argv[])
{
  
  struct soap v_soap;
  struct soap *v_tsoap; // clone of the SOAP context for the thread
  pthread_t tid;
  pid_t pid, sid;

  if( cmdline_parser (argc, argv, &ARGS) != 0)                                                                                                           
      exit(1) ;
  if(getConfiguration() < 0)                                                                                                                               
      {                                                                                                                                                    
	  debug( "Error while retrieving the configuration\n");
	  exit(1) ;
      }
    if( ! ARGS.nodaemon_given )
      {
         FILE *f;
         pid = fork();
         if (pid < 0)  exit(EXIT_FAILURE);
         if (pid > 0)  exit(EXIT_SUCCESS);
         umask(0);
         sid = setsid();
         if (sid < 0)  exit(1);
      
         close(STDIN_FILENO);
         close(STDOUT_FILENO);
         close(STDERR_FILENO); 
	
         f = fopen("/var/run/broker.pid", "w");
	 if(f != NULL) {
	     fprintf(f, "%d\n", getpid());
	     fclose(f);
	 } else {
	     f = fopen("/tmp/broker.pid", "w");
	     if(f != NULL) {
		 fprintf(f, "%d\n", getpid());
		 fclose(f);
	     }
	 }
     }
  setup_signals();
 
#ifndef NDEBUG
  debugClusters();
#endif
 
  if(acceptFrontalesConnections() < 0)
  {
    debug( "Error while waiting for frontales\n");
    return ERROR;
  }
  soap_init(&v_soap);
  soap_set_namespaces(&v_soap, service_namespaces);

  if(soap_bind(&v_soap, NULL, ARGS.port_arg, 100) < 0)
    return EXIT_FAILURE;

  while(running)
  {
    debug("Wating for requests ..... ");
    v_soap.accept_timeout = 60;  // non blocking
    if(soap_accept(&v_soap) < 0) 
      continue;
    v_tsoap=soap_copy(&v_soap);
    pthread_create(&tid, NULL, handleRequest, (void*)v_tsoap);
  }
  soap_done(&v_soap);
  return SUCCESS;
}
int delete_taskid (int taskid)
{
    int i,j;
    struct id *prev = NULL;
    struct id *curr = NULL;

    

    debug("Deleting task %d ..... ", taskid);
    debugClusters();
    i = 0;
    while (i < clusters_count)
        {
            j = 0;
            while(j <  clusters[i].hosts_count)
                {
                    prev = clusters[i].hosts[j].ids;
                    curr =  clusters[i].hosts[j].ids;
		    while((curr != NULL) && (curr -> process_name.taskid == taskid))
			{
			    clusters[i].hosts[j].ids = curr -> id_next;
			    free(curr);
			    curr = clusters[i].hosts[j].ids;
			}
		    curr =  clusters[i].hosts[j].ids;
		    if(curr != NULL)
			curr = curr -> id_next;
		    while(curr != NULL)
                        {
                            if(curr -> process_name.taskid == taskid)
                                {
                                    prev -> id_next = curr -> id_next;
				    free(curr);
				    curr = prev -> id_next;
				}
			    else
				{
				    prev = prev -> id_next;
				    curr = curr -> id_next;
				}
                        }
                    j++;
                }
            i++;
        }
    debugClusters();
    return SUCCESS;
}  
int  ns__delete_data(struct soap *p_soap, int *taskid)
{
    delete_taskid(*taskid);
    return SOAP_OK;  
}
int ns__service_proxy_port(struct soap *p_soap, int *service_proxy_port, struct ns__noResponse  *out)
{
  int i;

  for(i = 0; i < clusters_count; i++)
    clusters[i].proxy_service_port = *service_proxy_port;
  return SOAP_OK;
}
int ns__proxy_port(struct soap *p_soap, int *proxy_port, struct ns__noResponse  *out)
{
  int i;
  
  for(i = 0; i < clusters_count; i++)
    clusters[i].proxy_port = *proxy_port;
  return SOAP_OK;   
  
}

int ns__getHostURI(struct soap *p_soap, process_name_t *proc_name,
                   process_name_t *my_name, sockaddr_t *uri_A,
                   TAG_t tag, ns__uriResponse *uri_B)
{
  int found = 0;
  int i = 0, j = 0;
  int cluster_id_src = clusters_count;
  int cluster_id_dst = clusters_count;
  struct id *tmp = NULL;

  debug("ns__getHostURI TAG = %d %ld:%ld:%ld  %ld wants %ld:%ld:%ld %ld",
	tag, my_name->cellid, my_name->jobid, my_name->vpid,  my_name->taskid, 
	proc_name->cellid, proc_name->jobid, proc_name->vpid,  proc_name->taskid);
  debug("-----------------------");
  debugClusters();
  debug("-----------------------");

    found = 0;
  i = 0;
  while((found < 2) && (i < clusters_count))
    {
      j = 0;
      while((found < 2) && (j <  clusters[i].hosts_count))
	{
	    tmp = clusters[i].hosts[j].ids;
	    while(tmp != NULL)
		{
		    if((tmp->process_name.cellid == my_name->cellid) && 
		       (tmp->process_name.jobid == my_name->jobid) && 
		       (tmp->process_name.vpid == my_name->vpid) &&
		       (tmp->process_name.taskid == my_name->taskid))
			{
			    cluster_id_src = i;
			    found++;
			}
		    if( (tmp->process_name.cellid == proc_name->cellid) && 
			(tmp->process_name.jobid == proc_name->jobid) && 
			(tmp->process_name.vpid == proc_name->vpid) &&
			(tmp->process_name.taskid == my_name->taskid))
			{
			    cluster_id_dst = i;
			    uri_B -> host_name = (char *)malloc(MAX_HOST_NAME * sizeof(char));
			    memset(uri_B -> host_name, 0, MAX_HOST_NAME);
			    uri_B -> host_ip = (char *)malloc(MAX_IP * sizeof(char));
			    memset(uri_B ->host_ip, 0, MAX_IP);
			    if(tag == OMPI)
				{
				    debug("OMPI = %p %p %d ", tmp->addr_OMPI.host_ip, 
					  tmp->addr_OMPI.host_name,tmp->addr_OMPI.port);
				    strncpy(uri_B ->host_ip,tmp->addr_OMPI.host_ip, MAX_IP);
				    strncpy(uri_B ->host_name,tmp->addr_OMPI.host_name, MAX_HOST_NAME);
				    uri_B -> port = tmp->addr_OMPI.port;
				}
			    else
				{
				    debug("ORTE = %p %p %d ", tmp->addr_ORTE.host_ip, 
					  tmp->addr_ORTE.host_name,tmp->addr_ORTE.port);
				    strncpy(uri_B ->host_ip,tmp->addr_ORTE.host_ip, MAX_IP);
				    strncpy(uri_B ->host_name,tmp->addr_ORTE.host_name, MAX_HOST_NAME);
				    uri_B -> port = tmp->addr_ORTE.port;
				}
			    found++;
			}
		    tmp = tmp -> id_next;
		}
	    j++;
	}
      i++;
    }
  if(found < 2)
      {
	  debug( "Unknown host %s in ns__getMyURI\n", uri_B-> host_ip);
	  return SOAP_ERR;
      }
  
  debug("Result: ns__getHostURI TAG = %d %ld:%ld:%ld wants %ld:%ld:%ld  Cluster src = %d Cluster dst = %d",
        tag, my_name->cellid, my_name->jobid, my_name->vpid,
        proc_name->cellid, proc_name->jobid, proc_name->vpid,
	cluster_id_src,cluster_id_dst);
  
  if(cluster_id_dst == cluster_id_src)
      {
	  debug("-------ns__getMyURI same cluster-------");
	  uri_B -> traversing_code = direct;
	  uri_B -> response = NULL;
	  return SOAP_OK;
      }
  else
      {
	debug("-------ns__getMyURI not the same cluster ------- traversing method = %d", clusters[cluster_id_dst].traversing_code);
    switch (clusters[cluster_id_dst].traversing_code) 
    {
      case proxy:
        {
          struct soap jsoap;
          struct soap proxy_soap;
          int uid;
          char response[MAX_RESPONSE];
          char proxy_addr[MAX_ADDR];
          
          struct ns__noResponse out;
          
          soap_init(&proxy_soap);
          soap_set_namespaces(&proxy_soap, service_namespaces);
          sprintf(proxy_addr,"http://%s:%d/",clusters[cluster_id_src].proxy_ip, clusters[cluster_id_src].proxy_service_port);
          soap_call_ns__get_proxy_uid(&proxy_soap,proxy_addr,"",&uid);
          if(proxy_soap.error)
          {
            soap_print_fault(&proxy_soap, stderr);
            return SOAP_ERR;
          }
          soap_end(&proxy_soap);
          debug("UID from proxy %d",uid);
          
          uri_B -> traversing_code = proxy;
          uri_B -> response = (char *)malloc(MAX_RESPONSE * sizeof(char));
	  memset(uri_B -> response, 0, MAX_RESPONSE);

          sprintf(uri_B ->response, "%d", uid);
          strncpy(uri_B ->host_name, clusters[cluster_id_dst].proxy_name, MAX_HOST_NAME);
          strncpy(uri_B ->host_ip,clusters[cluster_id_dst].proxy_ip, MAX_IP);
          uri_B -> port = clusters[cluster_id_dst].proxy_port;
          
          soap_init(&jsoap);
          soap_set_namespaces(&jsoap, service_namespaces);
          pthread_mutex_lock(&frontale_lock);
          CONNECTED_SOCKET = clusters[cluster_id_dst].frontale_sockfd;
          jsoap.fconnect = doconnect;
          jsoap.fclose = doclose;
          sprintf(response,"%d.%d.%d.%d %d %s %d",(int)proc_name->cellid, (int)proc_name->jobid, (int)proc_name->vpid, tag, uid, 
          clusters[cluster_id_src].proxy_ip, clusters[cluster_id_src].proxy_port);
          soap_call_ns__proxy_connection_demande(&jsoap, "", "", response, &out);
          if(jsoap.error)
            soap_print_fault(&jsoap, stderr);
          pthread_mutex_unlock(&frontale_lock);
          debug("Sended to frontale %s",response);
          soap_destroy((struct soap*)&jsoap);
          soap_end((struct soap*)&jsoap);
          soap_done((struct soap*)&jsoap);
          return SOAP_OK;   
        }
        break;
    case direct:
      uri_B -> traversing_code = direct;
      uri_B -> response = NULL;
      return SOAP_OK;
    case tcp_traversing:
      uri_B -> traversing_code = tcp_traversing;
      uri_B -> response = NULL;
      return SOAP_OK;
      default:  
        debug("Traversing technques not yet implemented");
        break;
    }
	  return SOAP_ERR;
  }
  return SOAP_ERR;
}


int ns__bouchon(struct soap *p_soap, int p_code, char *p_chaine, struct ns__noResponse *p_out)
{
  debug("%s %d", p_chaine, p_code);
  return SOAP_OK;
}

int ns__forward_syn_data(struct soap *p_soap, char *data, struct ns__noResponse  *out)
{
    return SOAP_OK;
}
int ns__start_traversing(struct soap *p_soap, char *local_ip, int *port, struct ns__noResponse  *out)
{
  return SOAP_OK;
}

int ns__putMyURI(struct soap *p_soap, process_name_t *proc_name, 
		 sockaddr_t *uri_B, TAG_t tag, struct ns__noResponse *out)
{
    int found = 0, exists = 0;
  int i = 0, j = 0;
  struct id **tmp = NULL;
  char id[MAX_PROC_ID];

  sprintf(id,"%lu:%lu:%lu",proc_name->cellid, proc_name->jobid,proc_name->vpid);
   if( strcmp(id,"0:0:0") == 0)
      delete_taskid((int)proc_name->taskid);
    debug("ns__putMyURI TAG = %d '%s' %lu:%lu:%lu %lu", tag, uri_B->host_ip,proc_name->cellid, proc_name->jobid,proc_name->vpid, proc_name->taskid);
  
  while((!found) && (i < clusters_count))
    {
      j = 0;
      while((!found) && (j <  clusters[i].hosts_count))
	{
	  if((strcmp(clusters[i].hosts[j].host_name, uri_B -> host_ip) == 0) || 
	     (strcmp(clusters[i].hosts[j].host_name, uri_B -> host_name) == 0))
	    found = 1;
	  else
	    j++;
	}
      if(!found)
	i++;
    }
  if(!found)
    {
      debug( "Unknown host %s in ns__putMyURI\n", uri_B -> host_ip);
      return SOAP_ERR;
    }
  else
      debug( "Host found %s in ns__putMyURI\n", uri_B -> host_ip);  
  
  pthread_mutex_lock(&clusters_lock);
  
  debug("putMyUri clusters[i].hosts[j].ids = %p", clusters[i].hosts[j].ids);
  if( clusters[i].hosts[j].ids == NULL)
      {
	  clusters[i].hosts[j].ids =  malloc(sizeof(struct id));    
	  clusters[i].hosts[j].ids -> id_next = NULL;
	  tmp = &clusters[i].hosts[j].ids;
      }
  else
      {
	  struct id *next = clusters[i].hosts[j].ids;
	  
	  exists = 0;
	  
          while((!exists) && (next!=NULL))
              {
                  if((next->process_name.cellid == proc_name->cellid) &&
		     (next->process_name.jobid == proc_name->jobid) &&
		     (next->process_name.vpid == proc_name->vpid) &&
		     (next->process_name.taskid == proc_name->taskid))  
		      exists = 1;
		  else
		      next = next -> id_next; 
              }
	  if(!exists)
	      {
		  next = clusters[i].hosts[j].ids;
		  while(next -> id_next !=NULL)
		      next = next -> id_next;
		  next->id_next = malloc(sizeof(struct id));
		  (next->id_next) ->id_next = NULL;
		  tmp = &(next->id_next);   
	      }
	  else
	      tmp = &(next);
      }
  

  (*tmp)->process_name.cellid = proc_name->cellid;
  (*tmp)->process_name.jobid = proc_name->jobid;
  (*tmp)->process_name.vpid = proc_name->vpid;
  (*tmp)->process_name.taskid = proc_name->taskid;
  
  debugClusters();
  if(tag == OMPI)
    {
      (*tmp)->addr_OMPI.host_ip = (char *)malloc(MAX_IP * sizeof(char));
      (*tmp)->addr_OMPI.host_name = (char *)malloc(MAX_HOST_NAME * sizeof(char));

      strncpy((*tmp)->addr_OMPI.host_ip, uri_B -> host_ip, MAX_IP);
      strncpy((*tmp)->addr_OMPI.host_name, uri_B -> host_name, MAX_HOST_NAME);
      (*tmp)->addr_OMPI.port = uri_B -> port;
    }
  else
    {
     (*tmp)->addr_ORTE.host_ip = (char *)malloc(MAX_IP * sizeof(char));
     (*tmp)->addr_ORTE.host_name = (char *)malloc(MAX_HOST_NAME * sizeof(char));

      strncpy((*tmp)->addr_ORTE.host_ip, uri_B -> host_ip, MAX_IP);
      strncpy((*tmp)->addr_ORTE.host_name, uri_B -> host_name, MAX_HOST_NAME);
      (*tmp)->addr_ORTE.port = uri_B -> port;
    }
  pthread_mutex_unlock(&clusters_lock);
  return SOAP_OK;
}

static void* handleRequest(void *p_tsoap)
{
  struct soap *v_soap=(struct soap *)p_tsoap;
 
  pthread_detach(pthread_self());
  service_serve((struct soap*)v_soap);
  soap_destroy((struct soap*)v_soap);
  soap_end((struct soap*)v_soap);
  soap_done((struct soap*)v_soap);
  free(v_soap);
  pthread_exit(NULL);
}

int checkConfiguration(FILE *file)
{
  int i, c; 
  int id = 0, ncount = 0;
  char name[MAX_HOST_NAME];
  char proxy_name[MAX_HOST_NAME];
  char proxy_ip[MAX_IP];
  int portMin, portMax;
  
  TRAVERSING_CODE_t traversing_code;
  while((c = fscanf(file,"%d %d %d %s %s %s %d %d", &id, &ncount, (int *)&traversing_code, name, proxy_name, proxy_ip, &portMin, &portMax )) != EOF)
  {
    i = 0;
    clusters_count++;
    while((i < ncount)&&((c = fscanf(file,"%s",name)) != EOF))
      i++;
    if(c == EOF)
      return ERROR;
  }
  debug("NB of clusters = %d",clusters_count);
  return SUCCESS;
}

static void debugClusters()
{
  int i,j;
  struct id *tmp = NULL;
  
  for(i = 0; i < clusters_count; i++)
  {
    debug("Cluster id = %d frontale_sockfd = %d hosts_count = %d", clusters[i].id, clusters[i].frontale_sockfd, clusters[i].hosts_count);
    for(j = 0; j < clusters[i].hosts_count; j++)
    {
	debug("Host name = %s %p", clusters[i].hosts[j].host_name,clusters[i].hosts[j].ids);
      tmp =  clusters[i].hosts[j].ids;
      while(tmp != NULL)
	  {
	      debug("     %d         %ld:%ld:%ld",
		    tmp->process_name.taskid, 
		    tmp -> process_name.cellid,
		    tmp->process_name.jobid,
		    tmp->process_name.vpid); 
 	      tmp = tmp -> id_next; 
	  }
    }
  }
}

int freeClusters()
{
  int i,j;
  debug("Free Clusters");
  for(i = 0; i < clusters_count; i++)
  {
    for(j = 0; j < clusters[i].hosts_count; j++)
    {
	/*if(clusters[i].hosts[j].addr_OMPI.host_ip != NULL)
	  free(clusters[i].hosts[j].addr_OMPI.host_ip);
	  if(clusters[i].hosts[j].addr_OMPI.host_name != NULL)
	  free(clusters[i].hosts[j].addr_OMPI.host_name);
	  
	  if(clusters[i].hosts[j].addr_ORTE.host_ip != NULL)
	  free(clusters[i].hosts[j].addr_ORTE.host_ip);
	  if(clusters[i].hosts[j].addr_ORTE.host_name != NULL)
	  free(clusters[i].hosts[j].addr_ORTE.host_name);*/        
    }
    free(clusters[i].hosts);
  }
  free(clusters);
  clusters_count = 0;
  return SUCCESS;
}

int getConfiguration()
{
  FILE *file = NULL;
  int i,j,c,id,ncount;
  char name[MAX_HOST_NAME];
  char proxy_name[MAX_HOST_NAME];
  char proxy_ip[MAX_IP];
  int portMin, portMax;  

  TRAVERSING_CODE_t traversing_code;

  file = fopen(ARGS.conf_arg, "r");
  if(file == NULL)
  {
    perror("unable to open configuration file");
    return ERROR;
  }
  if(checkConfiguration(file) < 0)
  {
    debug( "configuration file %s is malformed\n", ARGS.conf_arg);
    fclose(file);
    return ERROR;
  }
  
  clusters = malloc(sizeof(struct cluster) * clusters_count);
  rewind(file);
  j = 0;
  while((c = fscanf(file,"%d %d %d %s %s %s %d %d", &id, &ncount, (int *)&traversing_code, name, proxy_name, proxy_ip, &portMin, &portMax)) != EOF)
  {   
    clusters[j].id = id;
    clusters[j].traversing_code = traversing_code;
    strncpy(clusters[j].frontale_name, name, MAX_HOST_NAME);
    strncpy(clusters[j].proxy_name, proxy_name, MAX_HOST_NAME);
    strncpy(clusters[j].proxy_ip, proxy_ip, MAX_IP);
    
    clusters[j].proxy_port = 0;
    clusters[j].proxy_service_port = 0;

    clusters[j].frontale_sockfd = 0;
    clusters[j].hosts_count = ncount;
    clusters[j].hosts = malloc(sizeof(struct host) * ncount);
    i = 0;
    while((i < ncount)&&((c = fscanf(file,"%s",name)) != EOF))
    {
      strncpy(clusters[j].hosts[i].host_name, name, MAX_HOST_NAME);
      clusters[j].hosts[i].ids  = NULL;
      i++;
    }
    j++;
  }
  return SUCCESS;
}

int putSocketById(int id, int client_sockfd)
{
  int i = 0, found = 0;
  while((!found) && (i < clusters_count))
  {
    debug("Cluster id = %d searching for %d", clusters[i].id, id);
    if(clusters[i].id == id)
    {
      found = 1;
      clusters[i].frontale_sockfd = client_sockfd;
    }
    i++;
  }
  if(found == 0)
    return ERROR;
  return SUCCESS;
}

int putSocketByNameOrIP(char *ip, char *name, int client_sockfd)
{
  int i = 0, found = 0;
  while((!found) && (i < clusters_count))
  {
    debug("Cluster frontale name = %s searching for %s or %s", clusters[i].frontale_name, ip, name);
    if((strcmp(clusters[i].frontale_name, ip) == 0) || (strcmp(clusters[i].frontale_name, name) == 0))
    {
      found = 1;
      strncpy(clusters[i].frontale_name,name,MAX_HOST_NAME);
      clusters[i].frontale_sockfd = client_sockfd;
      soap_init(&clusters[i].soap);
    }
    i++;
  }
  if(found == 0)
    return ERROR;
  return SUCCESS;
}
static int doconnect(struct soap *soap, const char *endpoint, const char *host, int port)
{
  printf("doconnect = %d\n", CONNECTED_SOCKET);
  soap->socket = CONNECTED_SOCKET;
  return SOAP_OK;
}
static int doclose(struct soap *soap)
{
  printf("doclose master = %d socket = %d \n",soap -> master, soap -> socket );
  return SOAP_OK;
}

int acceptFrontalesConnections()
{
  int sockfd,client_sockfd; 
  struct sockaddr_in serv_addr, client_addr; 
  int i = 0;
  socklen_t len = 0;
  
  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
  { 
    perror("server is unable to create a socket"); 
    return ERROR; 
  } 
  serv_addr.sin_family = AF_INET; 
  serv_addr.sin_addr.s_addr = INADDR_ANY; 
  serv_addr.sin_port = htons(ARGS.frontal_port_arg);
  
  if(bind(sockfd, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) < 0)
  {
    perror("unable to bind on port");
    return ERROR;
  }
  listen(sockfd, 5);  
  i = 0;
  while(i < clusters_count)
  {
    int nread;
    char buffer[MAX_BUFFER];
    char name[MAX_HOST_NAME];
    char ip[MAX_IP];
    
    debug("accepting next connection (opened connections: %d/%d)", i, clusters_count);
    
    len = sizeof(client_addr);
    if((client_sockfd = accept(sockfd, (struct sockaddr *)&client_addr, (socklen_t *)&len)) < 0)
    {
      perror("accept failed");
      return ERROR; 
    }
    
    nread = read(client_sockfd, buffer, MAX_BUFFER);
    
    if(nread <= 0)
    {
      perror("read failed");
      return ERROR;
    }
    sscanf(buffer, "%s %s", ip, name);
    if(putSocketByNameOrIP(ip, name, client_sockfd) < 0)
    {
      debug( "Unknown id %s %s for cluster\n", ip, name);
      close(client_sockfd);
      continue;
    }
    i++;
    
  }
  close(sockfd);
  return SUCCESS;
}

static void closeAllSockets()
{
  int i;
  for( i = 0; i< clusters_count; i++)
    close(clusters[i].frontale_sockfd);
}

void interrupt_catch(int signo)
{
  debug("interruption is caught");
  closeAllSockets();
  freeClusters();
  running = 0; 
  exit(1);     
}

void setup_signals(void)
{
  struct sigaction    sa;
  
  sa.sa_handler = SIG_IGN;                /* SIGPIPE to be ignored         */
  sigemptyset(&sa.sa_mask);               /* No signals to mask off        */
  sa.sa_flags = 0;                        /* No flags needed               */
 
  if (sigaction(SIGPIPE, &sa, NULL) == -1)
  {
    perror("sigaction(SIGPIPE)");
    exit(1);
  }
 
  sa.sa_handler = interrupt_catch;        /* SIGINT signal handler         */
  sigemptyset(&sa.sa_mask);               /* No signals to mask off        */
  sa.sa_flags = SA_RESTART;               /* Restart interrupted syscalls  */
 
  if (sigaction(SIGINT, &sa, NULL) == -1)
  {
    perror("sigaction(SIGINT)");
    exit(1);
  }
} 
int ns__update(struct soap *soap, int *number, struct ns__noResponse *out)
{
  return SOAP_OK;
}
int ns__update_broker(struct soap *soap, int *number, struct ns__noResponse  *out)
{
  printf("gsoap ns__update_broker frontale send me the number %d\n",*number);
  return SOAP_OK;
}
int ns__update_frontale(struct soap *soap, int *number, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int ns__update_connection_helper(struct soap *soap, int *number, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int ns__proxy_connection_demande(struct soap *soap, char *demande, struct ns__noResponse  *out)
{
  return SOAP_OK;
}

int ns__get_proxy_uid(struct soap *p_soap, int *uid)
{
  return SOAP_OK;
}

/***************
{
      struct soap jsoap;
      struct soap *tmpsoap;
      int number = 7;
      struct ns__noResponse out;
      
      soap_init(&jsoap);
      //soap_init2(&jsoap, SOAP_IO_KEEPALIVE, SOAP_IO_KEEPALIVE); 
      soap_set_namespaces(&jsoap, service_namespaces);
      tmpsoap = soap_copy(&jsoap);
       
      CONNECTED_SOCKET = client_sockfd;
      tmpsoap->fconnect = doconnect;
      tmpsoap->fclose = doclose;
      printf("gsoap socket %d\n", CONNECTED_SOCKET);
      
      soap_call_ns__update_frontale(tmpsoap, "http://129.175.7.55:6666/", "", &number, &out);
      if(tmpsoap->error)
      {
        soap_print_fault(tmpsoap, stderr);
        printf("soap error\n");
      }
      
      soap_destroy((struct soap*)tmpsoap);
      soap_end((struct soap*)tmpsoap);
      soap_done((struct soap*)tmpsoap);
      
      
      //soap_init(&jsoap);
      //soap_set_namespaces(&jsoap, service_namespaces);
      tmpsoap = soap_copy(&jsoap);
      soap_set_namespaces(tmpsoap, service_namespaces);
       number = 10;
      tmpsoap->fconnect = doconnect;
      tmpsoap->fclose = doclose;
      soap_call_ns__update_frontale(tmpsoap, "http://129.175.7.55:6666/", "", &number, &out);
      if(tmpsoap->error)
    {
       soap_print_fault(tmpsoap, stderr);
       printf("soap error\n");
    }
      soap_destroy((struct soap*)tmpsoap);
      soap_end((struct soap*)tmpsoap);
      soap_done((struct soap*)tmpsoap);
      
      soap_end(&jsoap);
    }
    **********/
    
    
    
