#ifndef MCA_BTL_SB_H
#define MCA_BTL_SB_H

/* Standard system includes */
#include "ompi_config.h"
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
/*#ifdef HAVE_SYS_SOCKET_H
  #include <sys/socket.h>
  #endif
  #ifdef HAVE_NETINET_IN_H
  #include <netinet/in.h>
  #endif */

/* Open MPI includes */
#include "opal/event/event.h"
#include "opal/util/output.h"
#include "ompi/class/ompi_bitmap.h"
#include "ompi/class/ompi_free_list.h"
#include "ompi/mca/pml/pml.h"
#include "ompi/mca/btl/btl.h"
#include "ompi/mca/btl/base/base.h"
#include "ompi/mca/mpool/mpool.h" 
#include "ompi/mca/btl/btl.h"

#if defined(c_plusplus) || defined(__cplusplus)
extern "C" {
#endif

  /**
   * Sender-based message-logging (SB) BTL component.
   */
  
  struct mca_btl_sb_component_t {
    mca_btl_base_component_1_0_1_t super;   /**< base BTL component */ 
    FILE *sb_fd;   /**< descriptor of the file used to log messages */
    
  } 
typedef struct mca_btl_sb_component_t mca_btl_sb_component_t;
OMPI_MODULE_DECLSPEC extern mca_btl_sb_component_t mca_btl_sb_component;

/**
 * BTL Module Interface
 */
struct mca_btl_sb_module_t {

}
typedef struct mca_btl_sb_module_t mca_btl_sb_module_t;
extern mca_btl_sb_module_t mca_btl_sb_module;


/**
 * Register SB component parameters with the MCA framework
 */
  extern int mca_btl_sb_component_open(void);
  
  /**
 * Any final cleanup before being unloaded.
 */
  extern int mca_btl_sb_component_close(void);
  
  /**
   * SB component initialization.
   * 
   * @param num_btl_modules (OUT)           Number of BTLs returned in BTL array.
   * @param allow_multi_user_threads (OUT)  Flag indicating wether BTL supports user threads (TRUE)
   * @param have_hidden_threads (OUT)       Flag indicating wether BTL uses threads (TRUE)
   */
  extern mca_btl_base_module_t** mca_btl_sb_component_init(
							   int *num_btl_modules, 
							   bool allow_multi_user_threads,
							   bool have_hidden_threads
							   );
  
  /**
   * SB component control.
   */
  int mca_btl_sb_component_control(
				    int param, 
				    void* value, 
				    size_t size
				    );
  
  
  /**
   * SB component progress.
   */
  extern int mca_btl_sb_component_progress(void);
  
  /**
   * Cleanup any resources held by the BTL.
   * 
   * @param btl  BTL instance.
   * @return     OMPI_SUCCESS or error status on failure.
   */
  
  extern int mca_btl_sb_finalize(
				  struct mca_btl_base_module_t* btl
				  );



#if defined(c_plusplus) || defined(__cplusplus)
}
#endif
#endif
